﻿using Dapper;
using EVCB_OCPP.TaskScheduler.Models;
using Quartz;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.Data.SqlClient;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Configuration;
using EVCB_OCPP.TaskScheduler.Helper;
using EVCB_OCPP.Domain;

namespace EVCB_OCPP.TaskScheduler.Jobs
{
    /// <summary>
    /// Update  EVSE's online status
    /// </summary>
    [DisallowConcurrentExecution]
    public class CheckEVSEOnlineJob : IJob
    {
        private readonly ILogger logger;
        private readonly SqlConnectionFactory<MainDBContext> mainDbConnectionFactory;
        private readonly SqlConnectionFactory<OnlineLogDBContext> onlineLogDbConnectionFactory;

        //private readonly string mainDBConnectString;
        //private readonly string onlineDBConnectString;

        private DateTime latestHeartbeatTime = DateTime.Now;
        private List<EVSEOnlineRecord> updateData = new List<EVSEOnlineRecord>();
        private List<EVSEOnlineRecord> insertData = new List<EVSEOnlineRecord>();

        public CheckEVSEOnlineJob(
            ILogger<CheckEVSEOnlineJob> logger,
            SqlConnectionFactory<MainDBContext> mainDbConnectionFactory,
            SqlConnectionFactory<OnlineLogDBContext> onlineLogDbConnectionFactory)
        {
            this.logger = logger;
            this.mainDbConnectionFactory = mainDbConnectionFactory;
            this.onlineLogDbConnectionFactory = onlineLogDbConnectionFactory;
            //this.mainDBConnectString = configuration.GetConnectionString("MainDBContext");
            //this.onlineDBConnectString = configuration.GetConnectionString("OnlineLogDBContext");
        }

        public async Task Execute(IJobExecutionContext context)
        {
            logger.LogDebug(this.ToString() + " :Starting........");
            try
            {
                List<EVSECurrentStatus> _EVSEs = await GetEVSEs();
                var checktime = DateTime.UtcNow.AddDays(-3);
                _EVSEs = _EVSEs.Where(x => x.HeartbeatUpdatedOn > checktime).ToList();

                foreach (var evse in _EVSEs)
                {

                    if (IsOnlineNow(evse))
                    {

                        if (!evse.Online)
                        { //off - on                           

                            await UpdateEVSECurrentStatus(evse.CustomerId.ToString(), evse.ChargeBoxId, true, DefaultSetting.DefaultNullTime);
                            await UpdateOnlineRecords(evse.ChargeBoxId, true, evse.HeartbeatUpdatedOn, null);
                        }
                    }
                    else
                    {

                        if (evse.Online)
                        {
                            //on -off                        

                            await UpdateEVSECurrentStatus(evse.CustomerId.ToString(), evse.ChargeBoxId, false, evse.HeartbeatUpdatedOn);
                            var online_row = await GetOnlineRecord(evse.ChargeBoxId);
                            if (online_row != null)
                            {
                                await UpdateOnlineRecords(evse.ChargeBoxId, false, evse.HeartbeatUpdatedOn, online_row.Id);
                            }

                        }

                    }
                }
            }
            catch (Exception ex)
            {
                logger.LogDebug("ERROR " + this.ToString() + ex.ToString());
            }



            logger.LogDebug(this.ToString() + " :Finished........");

        }


        private async Task<List<EVSECurrentStatus>> GetEVSEs()
        {
            List<EVSECurrentStatus> result = new List<EVSECurrentStatus>();
            try
            {
                using var dbConn = await mainDbConnectionFactory.CreateAsync();

                string sqlstring =
                    """
                    SELECT m.CustomerId,m.Id,m.ChargeBoxId,m.Online,m.HeartbeatUpdatedOn,MachineConfigurations.ConfigureSetting HeartbeatInterval
                    FROM [dbo].[Machine] m left join [dbo].[MachineConfigurations] MachineConfigurations  on m.ChargeBoxId = MachineConfigurations.ChargeBoxId
                    WHERE MachineConfigurations.ConfigureName = 'HeartbeatInterval'  and MachineConfigurations.ConfigureSetting!=''
                    """;
                result = (await dbConn.QueryAsync<EVSECurrentStatus>(sqlstring)).ToList();
            }
            catch (Exception ex)
            {
                logger.LogError("Query Data Error " + ex.ToString());
            }


            return result;
        }

        private async Task UpdateEVSECurrentStatus(string customerId, string ChargeBoxId, bool turnOn, DateTime offlineTime)
        {
            try
            {
                string sqlString = string.Format("UPDATE dbo.Machine SET Online=@Online {0} WHERE chargeBoxId=@chargeBoxId ", turnOn ? "" : " ,OfflineOn=@OfflineOn");
                using var dbConn = await mainDbConnectionFactory.CreateAsync();
                var parameters = new DynamicParameters();
                parameters.Add("@Online", turnOn, System.Data.DbType.Boolean);
                parameters.Add("@chargeBoxId", ChargeBoxId, System.Data.DbType.String);

                if (!turnOn)
                {
                    parameters.Add("@OfflineOn", offlineTime, System.Data.DbType.DateTime);
                }

                await dbConn.ExecuteAsync(sqlString, parameters);
            }
            catch (Exception ex)
            {
                logger.LogError("Update Data Error " + ex.ToString());
            }
        }


        private async Task UpdateOnlineRecords(string chargeBoxId, bool online, DateTime hearbeatDt, Int64? rowId)
        {
            try
            {
                if (online)
                {
                    string sqlString = "INSERT INTO dbo.EVSEOnlineRecord (\"ChargeBoxId\",\"OnlineTime\",\"OfflineTime\")" +
                       "VALUES( @ChargeBoxId,@OnlineTime,@OfflineTime); ";
                    using var dbConn = await onlineLogDbConnectionFactory.CreateAsync();
                    {
                        var parameters = new DynamicParameters();
                        parameters.Add("@ChargeBoxId", chargeBoxId, System.Data.DbType.String);
                        parameters.Add("@OnlineTime", hearbeatDt, System.Data.DbType.DateTime);
                        parameters.Add("@OfflineTime", DefaultSetting.DefaultNullTime, System.Data.DbType.DateTime);

                        await dbConn.ExecuteAsync(sqlString, parameters);
                    }
                }
                else
                {
                    if (rowId.HasValue)
                    {
                        string sqlString = "UPDATE dbo.EVSEOnlineRecord SET OfflineTime=@OfflineTime  WHERE Id=@Id";
                        using (var dbConn = await onlineLogDbConnectionFactory.CreateAsync())
                        {
                            var parameters = new DynamicParameters();
                            parameters.Add("@OfflineTime", hearbeatDt, System.Data.DbType.DateTime);
                            parameters.Add("@Id", rowId.Value, System.Data.DbType.Int64);

                            await dbConn.ExecuteAsync(sqlString, parameters);
                        }
                    }
                    else
                    {
                        logger.LogError("Can't find update row id from " + chargeBoxId);
                    }

                }
            }
            catch (Exception ex)
            {
                logger.LogError(string.Format("UpdateOnlineRecords Exception:{0}", ex.ToString()));
            }




        }



        async private Task<EVSEOnlineRecord> GetOnlineRecord(string chargeBoxId)
        {
            EVSEOnlineRecord lastrow = new EVSEOnlineRecord();
            try
            {
                string sqlString = string.Format("SELECT Id FROM dbo.EVSEOnlineRecord WHERE chargeBoxId=@chargeBoxId Order by OnlineTime desc");
                using (var dbConn = onlineLogDbConnectionFactory.Create())
                {
                    var parameters = new DynamicParameters();
                    parameters.Add("@chargeBoxId", chargeBoxId, System.Data.DbType.String);


                    var result = await dbConn.QueryAsync<EVSEOnlineRecord>(sqlString, parameters);
                    lastrow = result.FirstOrDefault();
                }

            }
            catch (Exception ex)
            {
                logger.LogError("Query Data Error " + ex.ToString());
            }
            return lastrow;
        }


        private bool IsOnlineNow(EVSECurrentStatus currentEVSE)
        {
            bool isOnline = false;
            int heartbeatInterval = 60;
            int.TryParse(currentEVSE.HeartbeatInterval, out heartbeatInterval);
            var checkTime = DateTime.UtcNow.AddSeconds(-(heartbeatInterval * 2));

            if (currentEVSE.HeartbeatUpdatedOn > checkTime)
            {
                isOnline = true;

            }

            return isOnline;

        }
    }
}
